
/****************************************************************************/
/*                                                                          */
/* 			     Module MEMORY                                  */
/* 			External Declarations 	                            */
/*                                                                          */
/****************************************************************************/

/* OSP constants */

#define MAX_PAGE       16                 /* max size of page tables        */
#define MAX_FRAME      32                 /* size of the physical memory    */
#define PAGE_SIZE      512                /* size of a page in bytes        */

#define COST_OF_PAGE_TRANSFER      6      /* cost of reading page from drum */

#define SET_SIZE       7	          /* size of working set 	    */

#define get_page_tbl(pcb)                 ( pcb->page_tbl )
#define MIN(a,b)                          ( a <= b ? a : b )

struct working_set_node {
   int pcb_id;			          /* pcb in question 	            */
   int top;			          /* pointer to LRU page 	    */
   int stack[SET_SIZE];		          /* working set 		    */
};

int snd_idx;

typedef struct working_set_node WORKING_SET;

/* external enumeration constants */

typedef enum {
    false, true                         /* the boolean data type            */
} BOOL;

typedef enum {
    read, write                         /* type of actions for I/O requests */
} IO_ACTION;

typedef enum {
    load, store                         /* types of memory reference        */
} REFER_ACTION;

typedef enum {
    running, ready, waiting, done       /* types of status                  */
} STATUS;

typedef enum {
    iosvc, devint,                      /* types of interrupt               */
    pagefault, startsvc,
    termsvc, killsvc,
    waitsvc, sigsvc, timeint
} INT_TYPE;


/* external type definitions */

typedef struct page_entry_node PAGE_ENTRY;
typedef struct page_tbl_node PAGE_TBL;
typedef struct event_node EVENT;
typedef struct ofile_node OFILE;
typedef struct pcb_node PCB;
typedef struct iorb_node IORB;
typedef struct int_vector_node INT_VECTOR;
typedef struct frame_node FRAME;


/* external data structures */

extern struct frame_node {
    BOOL   free;        /* = true, if free                                  */
    PCB    *pcb;        /* process to which the frame currently belongs     */
    int    page_id;     /* vitrual page id - an index to the PCB's page tbl */
    BOOL   dirty;       /* indicates if the frame has been modified         */
    int    lock_count;  /* number of locks set on page involved in an       */
                        /* active I/O                                       */
    int    *hook;       /* can hook up anything here                        */
};

extern struct page_entry_node {
    int    frame_id;    /* frame id holding this page                       */
    BOOL   valid;       /* page in main memory : valid = true; not : false  */
    int    *hook;       /* can hook up anything here                        */
};

extern struct page_tbl_node {
    PCB    *pcb;        /* PCB of the process in question                   */
    PAGE_ENTRY page_entry[MAX_PAGE];
    int    *hook;       /* can hook up anything here                        */
};

extern struct pcb_node {
    int    pcb_id;         /* PCB id                                        */
    int    size;           /* process size in bytes; assigned by SIMCORE    */
    int    creation_time;  /* assigned by SIMCORE                           */
    int    last_dispatch;  /* last time the process was dispatched          */
    int    last_cpuburst;  /* length of the previous cpu burst              */
    int    accumulated_cpu;/* accumulated CPU time                          */
    PAGE_TBL *page_tbl;    /* page table associated with the PCB            */
    STATUS status;         /* status of process                             */
    EVENT  *event;         /* event upon which process may be suspended     */
    int    priority;       /* user-defined priority; used for scheduling    */
    PCB    *next;          /* next pcb in whatever queue                    */
    PCB    *prev;          /* previous pcb in whatever queue                */
    int    *hook;          /* can hook up anything here                     */
};

extern struct iorb_node {
    int    iorb_id;     /* iorb id                                          */
    int    dev_id;      /* associated device; index into the device table   */
    IO_ACTION action;   /* read/write                                       */
    int    block_id;    /* block involved in the I/O                        */
    int    page_id;     /* buffer page in the main memory                   */
    PCB    *pcb;        /* PCB of the process that issued the request       */
    EVENT  *event;      /* event used to synchronize processes with I/O     */
    OFILE  *file;       /* associated entry in the open files table         */
    IORB   *next;       /* next iorb in the device queue                    */
    IORB   *prev;       /* previous iorb in the device queue                */
    int    *hook;       /* can hook up anything here                        */
};

extern struct int_vector_node {
    INT_TYPE cause;           /* cause of interrupt                         */
    PCB    *pcb;              /* PCB to be started (if startsvc) or pcb that*/
			      /* caused page fault (if fagefault interrupt) */
    int    page_id;           /* page causing pagefault                     */
    int    dev_id;            /* device causing devint                      */
    EVENT  *event;            /* event involved in waitsvc and sigsvc calls */
    IORB   *iorb;             /* IORB involved in iosvc call                */
};


/* extern variables */

extern INT_VECTOR Int_Vector;           /* interrupt vector         	     */
extern PAGE_TBL *PTBR;                  /* page table base register 	     */
extern FRAME Frame_Tbl[MAX_FRAME];      /* frame table              	     */
extern int Prepage_Degree;		/* global degree of prepaging (0-10) */


/* external routines */

extern siodrum(/* action, pcb, page_id, frame_id */);
/*  IO_ACTION   action;
    PCB         *pcb; 
    int         page_id, frame_id;  */
extern int get_clock();
extern gen_int_handler();

/****************************************************************************/

#define   PRIVATE         static
#define   PUBLIC
#define   NULL            0
#define   TRUE            1
#define   FALSE           0
#define   DELTA           10000      /* the delta for working set           */
#define   UNLOCKED        0
#define   NOT_REFERED     0
#define   NIL            -1          /* sorry! no free frame return         */
#define   UNDEFINED      -1          /* for PCB in FRAME and PAGE_TBL_ENTRY */
#define   END_OF_SET     -1          /* indicate end of working set         */
#define   MAX_SIZE        MAX_PAGE*PAGE_SIZE /* max size of a job allowed   */
#define   SET_SIZE        7          /* size of working set                 */
#define   MAX_PCB         50         /* max num of PCBs in Pcb_pool         */

/* internal data structure for memory.c */

static WORKING_SET Working_Set_Pool[MAX_PCB];

#define FRACTION 0.1 * Prepage_Degree

static int trace = TRUE;
int time_stamps[MAX_FRAME];

WORKING_SET *find_working_set();

/**************************************************************************/
/*									  */
/*			memory_init()					  */
/*									  */
/*									  */
/*   called by    : SIMCORE module					  */
/*									  */
/**************************************************************************/

memory_init()
{
     int i , j;

     /* Initialize free frame list */
     for ( i = 0 ; i < MAX_FRAME ; ++i ) {
        Frame_Tbl[i].pcb        = 0;     /* pcb to which the frame belongs  */
        Frame_Tbl[i].page_id    = 0;     /* index to the PCB's page tbl     */
        Frame_Tbl[i].lock_count = 0;     /* # of locks set                  */
        Frame_Tbl[i].dirty      = false; /* if the frame has been modified  */
        Frame_Tbl[i].free       = true;  /* = true , if free                */
        time_stamps[i]          = get_clock();
        /* make reference bit for second chance replacement algorithm       */
        Frame_Tbl[i].hook       = ( int * ) malloc ( sizeof(int) );
        *(Frame_Tbl[i].hook)    = false; /* not referenced                  */
     }

     /* Initialize working set pool */
     for ( i = 0 ; i < MAX_PCB ; ++i ) {
        Working_Set_Pool[i].top    = NIL;
        Working_Set_Pool[i].pcb_id = NIL;
        for ( j = 0 ; j < SET_SIZE ; ++j ) 
           Working_Set_Pool[i].stack[j] = 0;
     }

     /* Initialize index for second chance replace */
     snd_idx = 0;
}

/*************************************************************************/
/*                                                                       */
/*   set_working_set                                                     */
/*                                                                       */
/*   set  the next available working set from the working set pool.      */
/*                                                                       */
/*************************************************************************/

PRIVATE
set_working_set(page_tbl)
PAGE_TBL *page_tbl;
{
     int i;

     i = 0;
     while ( ( i < MAX_PCB ) && ( Working_Set_Pool[i].pcb_id != NIL ) )
        i++;

     if ( i == MAX_PCB ) {
        printf("CLOCK> %6d#*** MEMORY> OSP Bug in working set handling\n",
                                        get_clock());
		    /* this should not happen, since there are as many entries
		      in the working set pool as there are active processes. */
        osp_abort();
     }

     Working_Set_Pool[i].pcb_id = page_tbl->pcb->pcb_id;
}

/*************************************************************************/
/*									 */
/*			set_frame                       		 */
/*								         */
/*    	Description : To set the memory frame to the input page number   */
/*                    and pcb                                            */
/*									 */
/*      Called by   : get_page()					 */
/*								         */
/*************************************************************************/

PRIVATE
set_frame(pcb,page_no,frame_no)
PCB *pcb;
int page_no;
int frame_no;
{
     int      page_id;
     PAGE_TBL *page_tbl_ptr;

     printf("set_frame3\n");  

     if ( trace )
        printf("Hello set_frame(pcb %d,page_no %d,frame_no %d)\n",
			pcb->pcb_id,page_no,frame_no);

     if ( Frame_Tbl[frame_no].pcb != NULL ) {
        page_tbl_ptr = get_page_tbl(Frame_Tbl[frame_no].pcb);	/* macro */
        page_id      = Frame_Tbl[frame_no].page_id;
        page_tbl_ptr->page_entry[page_id].valid = false;
     } 

     Frame_Tbl[frame_no].pcb        = pcb;
     Frame_Tbl[frame_no].page_id    = page_no;
     Frame_Tbl[frame_no].lock_count = UNLOCKED;
     Frame_Tbl[frame_no].dirty      = false;
     Frame_Tbl[frame_no].free       = false;
     time_stamps[frame_no]          = get_clock();
}

/************************************************************************/
/*		    mem_find_frame   					*/
/*                                                                      */
/*	description: Find a frame from the Frame_Tbl according to the   */
/*                   following criteria:                                */
/*                                                                      */
/*                   1. frame must be unlocked                          */
/*                                                                      */
/*                   2. pick the free frame (pcb == NULL)first          */
/*                                                                      */
/*                   3. pick the LRU frame allocated to other process   */
/*									*/
/************************************************************************/

PRIVATE
int mem_find_frame(pcb)
PCB *pcb;
{
     int old_idx , i;
  
     /*
      * serach frame list
      *	if there is free frame allocate it to the pcb;
      */

     for ( i = 0 ; i < MAX_FRAME ; ++i ) { 
	if ( Frame_Tbl[i].free && Frame_Tbl[i].lock_count == UNLOCKED ) {
	     return(i); /* got free frame */
	}
     }

     /*
      * find victim using second chance algorithm;
      * allocate the replaced frame to the pcb;
      * global replacement used;
      */

     old_idx = snd_idx;

     while ( true ) {
	for ( snd_idx ; snd_idx < MAX_FRAME ; ++snd_idx ) {
	   if ( *(Frame_Tbl[snd_idx].hook) == NOT_REFERED && 
	        Frame_Tbl[snd_idx].lock_count == UNLOCKED ) {
	        return(snd_idx); /* victim found    */
	   }
	   else {                /* give 2nd chance */
		*(Frame_Tbl[snd_idx].hook) = false;
	   }
	}
	for ( snd_idx = 0 ; snd_idx < old_idx ; ++snd_idx ) {
	   if ( *(Frame_Tbl[snd_idx].hook) == NOT_REFERED  &&
	        Frame_Tbl[snd_idx].lock_count == UNLOCKED ) {
	   	return(snd_idx); /* victim found    */
	   }
	   else {                /* give 2nd chance */
		*(Frame_Tbl[snd_idx].hook) = false;
	   }
	}
     }			
}

/**************************************************************************/
/*									  */
/*			get_page					  */
/*									  */
/*									  */
/**************************************************************************/

get_page(pcb,page_id)
PCB *pcb;
int page_id;
{
     int victim;

     victim = mem_find_frame(pcb);
      if ( Frame_Tbl[victim].dirty == true ) 
          siodrum(write,Frame_Tbl[victim].pcb,Frame_Tbl[victim].page_id,victim);
     set_frame(pcb,page_id,victim);
     siodrum(read,pcb,page_id,victim);
     pcb->page_tbl->page_entry[page_id].valid = true;
     pcb->page_tbl->page_entry[page_id].frame_id = victim;
}

/*************************************************************************/
/*                                                                       */
/*   clear_working_set                                                   */
/*                                                                       */
/*                                                                       */
/*************************************************************************/ 

clear_working_set(page_tbl)
PAGE_TBL *page_tbl;
{
     WORKING_SET *ws;
     int i;

     ws = find_working_set(page_tbl); /* find working set of current process */
     if ( ws != NULL ) {              /* if working set is found */
        ws->top    = NIL;
        ws->pcb_id = NIL;
        for ( i = 0 ; i < SET_SIZE ; ++i ) 
              ws->stack[i] = 0;
     }
}

/**************************************************************************/
/*									  */
/*                  deallocate						  */
/*   The job is history now so free the memory frames occupied by process */
/*   Just set the PCB to NULL.                                            */
/*									  */
/**************************************************************************/

deallocate(pcb)
PCB *pcb;
{
     int frame_id , i;
     PAGE_TBL *temp;

     temp = pcb->page_tbl;
     for ( i = 0 ; i < MAX_PAGE ; i++ ) {
	/* if there is an entry of the page that is if valid bit is true */
	if ( temp->page_entry[i].valid == true ) {
	/* free the memory as follows */
	/* reset pcb, free & dirty fields for Frame_Tbl[frame_id]; */
	   frame_id = temp->page_entry[i].frame_id;
	   Frame_Tbl[frame_id].pcb   = 0; 
	   Frame_Tbl[frame_id].free  = true;
	   Frame_Tbl[frame_id].dirty = false;
	   *(Frame_Tbl[frame_id].hook) = false;
           temp->page_entry[i].valid = false;
	}
     }
     clear_working_set(temp);
     clear_page_tbl_entry(temp);
}

/************************************************************************/
/*									*/
/*		         initial_swap_in            			*/
/*								        */
/*	Description :  Swap number of page in the main            .     */
/*									*/
/*									*/
/*     called by : swap_in()					        */
/*     call      : get_page()					        */
/*                                                                      */
/************************************************************************/

PRIVATE
initial_swap_in(num_of_pages,pcb)
int num_of_pages;
PCB *pcb;
{
     int i;

     if ( trace )
       printf("Hello initial_swap_in(num_of_pages: %d,pcb:%d)\n",
			num_of_pages,pcb->pcb_id);

     i = 0;
     while ( i < MAX_PAGE && num_of_pages > 0 ) {
       get_page(pcb,i);
       num_of_pages--;
       i++;
     }
}

/************************************************************************/
/*									*/
/*		         swap_in_working_set           			*/
/*								        */
/*	Description :  Swap the working set into main memory            */
/*									*/
/*									*/
/*     called by : swap_in()					        */
/*     call      : get_page()					        */
/*                                                                      */
/************************************************************************/

PRIVATE
swap_in_working_set(working_set,pcb)
WORKING_SET *working_set;
PCB  *pcb;
{
     int i;
     int page_num;

     if ( trace )
        printf("Hello swap_in_working_set( working_set: %d,pcb:%d)\n",
		working_set ,pcb->pcb_id);

     if ( working_set->top != NIL ) { /* working set is nonempty */
        i = working_set->top;
        page_num = ( working_set->top ) * FRACTION;

        while ( ( i-- >= 0 ) && ( page_num-- > 0 ) ) { /* from the top */
           get_page(pcb,working_set->stack[i]);
        }
     }
}

/************************************************************************/
/*									*/
/*	  clear_page_tbl_entry()					*/
/*									*/
/************************************************************************/

PRIVATE
clear_page_tbl_entry(page_tbl_ptr)
PAGE_TBL *page_tbl_ptr;
{
     int i;

     for ( i = 0 ; i < MAX_PAGE ; ++i ) {
	page_tbl_ptr->page_entry[i].valid = false;
     }
}
					 
/************************************************************************/
/*									*/
/*		   prepage						*/
/*									*/
/*									*/
/************************************************************************/

PUBLIC
prepage(pcb)
PCB *pcb;
{
     PAGE_TBL *page_table_ptr;
     WORKING_SET *w_s;
     int swap_in_size , no_of_pages;

    /*
     * swap the process specified in the argument from drum/disk to main 
     * memory.
     * will use the prepaging policy.
     */

     page_table_ptr = pcb->page_tbl;
     page_table_ptr->pcb = pcb;

     if ( ( w_s = find_working_set(page_table_ptr) ) )
        swap_in_working_set(w_s,pcb);
     else
     {
        swap_in_size = MIN(pcb->size,SET_SIZE*PAGE_SIZE) * FRACTION;
	no_of_pages = swap_in_size / PAGE_SIZE;
        if ( swap_in_size % PAGE_SIZE ) no_of_pages++;
        initial_swap_in(no_of_pages,pcb);
        set_working_set(page_table_ptr);
     }	
}   
					 
/************************************************************************/
/*									*/
/*		   start_cost						*/
/*									*/
/************************************************************************/

int start_cost(pcb)
PCB *pcb;
{
     int i , swap_in_size , num_of_pages , page_in;
     PAGE_TBL *page_table_ptr;
     WORKING_SET *ws;

     page_in = 0;  
     page_table_ptr = pcb->page_tbl;
     ws = find_working_set(page_table_ptr);
     if ( ws != NULL && ws->top != NIL ) {
       num_of_pages = ( ws->top + 1 ) * FRACTION;
       for ( i = 0 ; i < num_of_pages ; ++i ) {
	 if ( page_table_ptr->page_entry[ws->top-i].valid == false )
	   page_in++;
       }
     }
     else {
       swap_in_size = MIN(pcb->size,SET_SIZE*PAGE_SIZE) * FRACTION;
       num_of_pages = swap_in_size / PAGE_SIZE;
       if ( swap_in_size % PAGE_SIZE ) num_of_pages++;
       for ( i = 0 ; i < num_of_pages ; i++ ) {
	 if ( page_table_ptr->page_entry[i].valid == false )
	   page_in++;
       }
     }
     return(COST_OF_PAGE_TRANSFER*page_in);
}   

/***************************************************************************/
/*                                                                         */
/*    update_working_set                                                   */
/*                                                                         */
/*    Note: top always points to the LRU page no                           */
/*                                                                         */
/***************************************************************************/

PRIVATE
update_working_set(working_set,page_no)
WORKING_SET *working_set;
int         page_no;
{
     int i , page_num;

     /* search stack for same page and change position in the stack */
     for ( i = 0 ; i <= working_set->top ; ++i ) {
	if ( working_set->stack[i] == page_no ) { /* already exist  */
	     page_num = working_set->stack[i];    /* save for use   */
	     for ( ; i < working_set->top ; ++i ) {
		working_set->stack[i] = working_set->stack[i+1];
	     }
	     working_set->stack[working_set->top] = page_num;
	     return;
	}
     }

     /* push new page number into the stack */
     if ( working_set->top < SET_SIZE - 1 ) {
        working_set->top++;
        working_set->stack[working_set->top] = page_no;
     }
     else {
        /* top == SET_SIZE - 1, i.e. stack is full  */
        /* squeeze out the bottom page and shift the upper pages down  */

        for ( i = 0 ; i < working_set->top; i++ ) 
           working_set->stack[i] = working_set->stack[i+1];

        working_set->stack[working_set->top] = page_no;
     }
}

/****************************************/
/* finds working set for the page table */
/****************************************/

WORKING_SET *find_working_set(page_tbl)
PAGE_TBL *page_tbl;
{
     int i;

     for ( i = 0 ; i <= MAX_PCB ; i++ ) {
	if ( Working_Set_Pool[i].pcb_id == page_tbl->pcb->pcb_id )
	   return(&Working_Set_Pool[i]);
     }
     return(NULL);
} 
 
/***************************************************************************/
/*									   */
/*   			refer					           */
/*									   */
/*									   */
/***************************************************************************/

refer(logic_addr,action)
int          logic_addr;   /* logical address                              */
REFER_ACTION action;       /* a store operation will change memory content */
{
     int job_size , page_no , frame_id;
     PAGE_TBL *tbl_ptr;
     PCB *pcb;
     WORKING_SET *ws;

     pcb      = PTBR->pcb;          /* current PCB pointer */
     job_size = pcb->size;          /* current job size */
     tbl_ptr  = pcb->page_tbl;      /* page table pointer of current PCB */

     if ( logic_addr < MAX_SIZE ) { /* max size of a job allowed */
       if ( logic_addr >= 0 && logic_addr < job_size ) {
           page_no = logic_addr / PAGE_SIZE; 
           /* if the page # is not in memory , generate page fault interrupt */
           if ( tbl_ptr->page_entry[page_no].valid == false ) {
	        Int_Vector.cause   = pagefault;
		Int_Vector.pcb     = pcb;
		Int_Vector.page_id = page_no;
	        gen_int_handler();
           }

           /* update working set */
           ws = find_working_set(tbl_ptr);
           if ( ws == NULL ) {
        	if ( Prepage_Degree > 0 ) {
                     printf("CLOCK> %6d#*** MEMORY> Bug in WS\n",get_clock());
             	     osp_abort();
	        }
	        else {
	             set_working_set(tbl_ptr);
	             ws = find_working_set(tbl_ptr);
	        }
           }
           update_working_set(ws,page_no);
           frame_id = tbl_ptr->page_entry[page_no].frame_id;
	   *(Frame_Tbl[frame_id].hook) = true; /* reference bit on */
           if ( action == store ) {
		Frame_Tbl[frame_id].dirty = true; 
           }
       }
     }
} 

/*************************************************************************/
/*									 */
/*			lock_page					 */
/*									 */
/*************************************************************************/

lock_page(iorb)
IORB *iorb;
{
     int page_id , frame_id;
     PAGE_TBL *tbl_ptr;

     tbl_ptr = iorb->pcb->page_tbl;
     page_id = iorb->page_id;

     if ( tbl_ptr->page_entry[page_id].valid == false ) { /* not in memory */
       Int_Vector.cause   = pagefault;
       Int_Vector.pcb     = iorb->pcb;
       Int_Vector.page_id = page_id;
       gen_int_handler();
     }
     frame_id = tbl_ptr->page_entry[page_id].frame_id;
     if ( iorb->action == read ) { /* from disk to main memory */
       Frame_Tbl[frame_id].dirty = true;
     }
     Frame_Tbl[frame_id].lock_count++;
} 

/*************************************************************************/
/*                                                                       */
/*              unlock_page()                                            */
/*                                                                       */
/*************************************************************************/

unlock_page(iorb)
IORB *iorb;
{
     int page_id , frame_id;
     PAGE_TBL *tbl_ptr;

     tbl_ptr = iorb->pcb->page_tbl;
     page_id = iorb->page_id;
     frame_id = tbl_ptr->page_entry[page_id].frame_id;
     Frame_Tbl[frame_id].lock_count--;
} 

/****************************************************************************/
/*									    */
/*			print_page_tbl					    */
/*									    */
/****************************************************************************/

PRIVATE
print_workset_pool()

{
     int i,j;

     for ( j = 0 ; j < MAX_PCB ; j++ ) {
	printf("Work set of PCB %d  = ", Working_Set_Pool[j].pcb_id);
	for ( i = 0 ; i <= Working_Set_Pool[j].top ; i++ )
		printf(" %d", Working_Set_Pool[j].stack[i]);
	printf("\n\n");
     }
}
